from vcScript import *
from vcHelpers.Selection import *

finalized = False
finger_node = None
finger_joint = None
grasping_signal = None

#Register to the OnNodeTransform event
def OnStart():
  global finger_node, finger_joint
  if control_mode_prop is not None and control_mode_prop.Value != "Emulation":
    return
  finger_node = get_finger_node()
  if finger_node is not None and enabled_prop.Value:
    finger_joint = finger_node.Dof
    finger_node.OnNodeTransform = handler

#Unregister to the OnNodeTransform event
def OnReset():
  global finger_node
  if finger_node is not None:
    finger_node.OnNodeTransform = None

#return the node that should trigger the grasping
def get_finger_node():
  if trigger_joint_prop.Value is None:
    return None
  split_name = trigger_joint_prop.Value.split(":")
  split_name = split_name[0].strip()
  finger_node = comp.findNode(split_name)
  return finger_node

#Main event handler for triggering the grasp signal
def handler(arg):
  if not sim.IsRunning:
    return
  if is_closed(finger_joint):
    if grasping_signal.Value != True:
      grasping_signal.signal(True)
  elif is_open(finger_joint):
    if grasping_signal.Value != False:
      grasping_signal.signal(False)

#Return difference between current joint value and the trigger point
def get_delta(joint):
  return joint.VALUE - trigger_point_prop.Value

def is_closed(joint):
  #return abs(abs(finger_joint.VALUE) - CLOSED_VAL) < EPSILON
  if invert_prop.Value:
    return get_delta(joint) < trigger_delta_prop.Value
  return get_delta(joint) > (-trigger_delta_prop.Value)

def is_open(joint):
  #return abs(abs(finger_joint.VALUE) - OPEN_VAL) < EPSILON
  if invert_prop.Value:
    return get_delta(joint) > (-trigger_delta_prop.Value)
  return get_delta(joint) < trigger_delta_prop.Value

#Update the trigger joint property step values when the comp nodes are changed (new node might have been added)
def update_trigger_options(type_constant, node1, node2):
  if type_constant == VC_NODE_REMOVE_NODE and node1 == comp: #Component deleted
    return
  elif type_constant != VC_NODE_REMOVE_NODE: #Nodes were modified
    set_trigger_servo(node1)
  else: #Node other than the component was removed
    set_trigger_servo(comp)

#Executed when the component is loaded
def OnFinalize():
  global grasping_signal, finalized
  if finalized:
    return
  finalized = True
  grasping_signal = grasping_node.getBehaviour(grasping_node.Name + " Signal")
  if grasping_signal is None:
    print "Error. Failed to find the grasping signal"
  get_trigger_joint()

#Initializes the TriggerJoint value and step constraints
def get_trigger_joint():
  parent_node = grasping_node.Parent
  finger_node = set_trigger_servo(parent_node)
  if finger_node is None:
    print "Failed to set the TriggerJoint"

#Initialize Trigger Servo to the other child of the grasping node
def set_trigger_servo(selected_node):
  children = selected_node.Children
  all_children = getGivenComponentsNodes(comp)
  #Combine the two lists without duplicates
  children = list(children)
  children.extend(x for x in all_children if x not in children)
  if None in children:
    return None
  nodes_and_dofs, trigger_node = build_trigger_step_values(children)
  if not nodes_and_dofs:
    print("No valid servos found for Trigger Servo")
    return None
  trigger_joint_prop.StepValues = nodes_and_dofs
  if trigger_joint_prop.Value is None:
    trigger_joint_prop.Value = trigger_joint_prop.StepValues[0]
  return trigger_node

#Return a list of Node name & Dof name, and the first trigger node to be the default value
def build_trigger_step_values(children):
  trigger_node = None
  nodes_and_dofs = []
  VALID_JOINT_TYPES = [VC_JOINTTYPE_TRANSLATIONAL, VC_JOINTTYPE_ROTATIONAL, VC_JOINTTYPE_CUSTOM]
  for c in children:
    if c.Dof is None or c.Dof.JointLinkType not in VALID_JOINT_TYPES:
      continue
    if trigger_node == None:
      trigger_node = c
    combined_names = "{} : {}".format(c.Name, c.Dof.Name)
    nodes_and_dofs.append(combined_names)
  return nodes_and_dofs, trigger_node

#Validate that the given property exists, and return the property
def confirmProperty(name,defaultvalue,type=VC_REAL,constraints=VC_PROPERTY_DEFAULT):
     prop = comp.getProperty(name)
     if type == VC_VECTOR:
        if prop == None:
          prop = comp.createProperty(type,name,constraints)
          prop.Value = default_detection_volume
     else:
        if prop == None:
          prop = comp.createProperty(type,name,constraints)
          if constraints == VC_PROPERTY_STEP:
            prop.StepValues = DEFAULT_TRIGGER_JOINT_VALS
          prop.Value = defaultvalue
     return prop

#hide/show the trigger properties if the trigger is enabled or not
def toggle_hide_trigger_props(arg):
  trigger_joint_prop.IsVisible = arg.Value
  trigger_point_prop.IsVisible = arg.Value
  trigger_delta_prop.IsVisible = arg.Value
  invert_prop.IsVisible = arg.Value


comp = getComponent()
sim = getSimulation()
grasping_node = getNode()

DEFAULT_EPSILON = 0.5
DEFAULT_TRIGGER_POINT = 5.0
DEFAULT_TRIGGER_JOINT_VALS = []
DEFAULT_TRIGGER_JOINT = ''
enabled_prop = confirmProperty(grasping_node.Name + "::ServoTriggerEnabled", True, VC_BOOLEAN)
trigger_joint_prop = confirmProperty(grasping_node.Name + "::TriggerJoint", DEFAULT_TRIGGER_JOINT, VC_STRING, VC_PROPERTY_STEP)
trigger_point_prop = confirmProperty(grasping_node.Name + "::TriggerPoint", DEFAULT_TRIGGER_POINT, VC_REAL)
trigger_delta_prop = confirmProperty(grasping_node.Name + "::TriggerDelta", DEFAULT_EPSILON, VC_REAL)
invert_prop = confirmProperty(grasping_node.Name + "::InvertTrigger", False, VC_BOOLEAN)

#Project specific property
control_mode_prop = comp.getProperty("ControlMode")

enabled_prop.OnChanged = toggle_hide_trigger_props
comp.OnNodeConfigurationChange = update_trigger_options
OnFinalize()
